import org.jetbrains.exposed.dao.*
import org.jetbrains.exposed.dao.id.EntityID
import org.jetbrains.exposed.dao.id.IntIdTable
import org.jetbrains.exposed.sql.*
import org.jetbrains.exposed.sql.transactions.transaction
import java.sql.SQLException
import java.util.*

// Teil 1: Singletons für die Tabellen (Plural)
object Publishers : IntIdTable() {
    val name = varchar("name", length = 100)
}
object Books : IntIdTable() {
    val title = varchar("title", 250)
    val publyear = integer("publyear")
    val publId = reference("publId", Publishers)
}
object Authors : IntIdTable() {
    val name = varchar("name", 100)
}
object BookAuthors : Table() {
    val bookId = reference("bookId", Books)
    val authorId = reference("AuthorId", Authors)
    override val primaryKey = PrimaryKey(bookId, authorId)
}

// Teil 2: Klassen für Datensätze (Singular)
class Publisher(id: EntityID<Int>) : IntEntity(id) {
    companion object : IntEntityClass<Publisher>(Publishers)
    var name by Publishers.name
    // books: Datentyp SizedIterable<Book>
    val books by Book referrersOn Books.publId
}
class Book(id: EntityID<Int>) : IntEntity(id) {
    companion object : IntEntityClass<Book>(Books)
    var title by Books.title
    var publyear by Books.publyear
    var publId by Books.publId
    // publishers: Datentyp Publisher
    val publisher by Publisher referencedOn Books.publId
    // authors: Datentyp SizedIterable<Author>
    val authors by Author via BookAuthors

}
class Author(id: EntityID<Int>) : IntEntity(id) {
    companion object : IntEntityClass<Author>(Authors)
    var name by Authors.name
    // books: Datentyp SizedIterable<Book>
    val books by Book via BookAuthors
}

// Teil 3: Testcode
fun setupOneBook() {
    transaction {
        // Tabellen erzeugen
        for (t in listOf(Publishers, Authors, Books, BookAuthors))
            SchemaUtils.create(t)

        // nur ein einzelnes Buch speichern
        val rwid = Publishers.insert {
            it[name] = "Rheinwerk Verlag"
        }[Publishers.id]
        val mkid = Authors.insert {
            it[name] = "Kofler, Michael"
        }[Authors.id]
        val bkid = Books.insert {
            it[title] = "Kotlin -- Das umfassende Handbuch"
            it[publyear] = 2020
            it[publId] = rwid
        }[Books.id]
        BookAuthors.insert {
            it[bookId] = bkid
            it[authorId] = mkid
        }

        // testen, ob Verlag schon existiert; ggfs. einfügen
        val newid1 = findOrInsertPublisher("testverlag")
        val newid2 = findOrInsertPublisher("testverlag")
        println("id1=$newid1 id2=$newid2")
    }
}

fun setupBooks() {
    transaction {
        // Logging
        // addLogger(StdOutSqlLogger)

        // Tabellen erzeugen
        for (t in listOf(Publishers, Authors, Books, BookAuthors))
            SchemaUtils.createMissingTablesAndColumns(t)

        // wenn es schon Datensätze gibt: abbrechen
        if (Publishers.selectAll().count() > 0)
                return@transaction

        // Verlage
        val publishers = listOf("ebooks.kofler", "Rheinwerk Verlag")
        val publIds = mutableMapOf<String, EntityID<Int>>()
        for (pname in publishers)
            publIds[pname] = Publishers.insert {
                it[name] = pname
            }[Publishers.id]

        // Autoren
        val authors = listOf(
            "Kofler, Michael", "Öggl, Bernd",
            "Kühnast, Charly", "Scherbeck, Christoph"
        )
        val authIds = mutableMapOf<String, EntityID<Int>>()
        for (aname in authors)
            authIds[aname] = Authors.insert {
                it[name] = aname
            }[Authors.id]

        // Kotlin-Buch
        var bkid = Books.insert {
            it[title] = "Kotlin -- Das umfassende Handbuch"
            it[publyear] = 2020
            it[publId] = publIds["Rheinwerk Verlag"]!!
        }[Books.id]
        BookAuthors.insert {
            it[bookId] = bkid
            it[authorId] = authIds["Kofler, Michael"]!!
        }

        // Linux-Buch
        bkid = Books.insert {
            it[title] = "Linux -- Das umfassende Handbuch"
            it[publyear] = 2019
            it[publId] = publIds["Rheinwerk Verlag"]!!
        }[Books.id]
        BookAuthors.insert {
            it[bookId] = bkid
            it[authorId] = authIds["Kofler, Michael"]!!
        }

        // Raspberry-Pi-Buch
        bkid = Books.insert {
            it[title] = "Raspberry Pi -- Das umfassende Handbuch"
            it[publyear] = 2019
            it[publId] = publIds["Rheinwerk Verlag"]!!
        }[Books.id]
        BookAuthors.insert {
            it[bookId] = bkid
            it[authorId] = authIds["Kofler, Michael"]!!
        }
        BookAuthors.insert {
            it[bookId] = bkid
            it[authorId] = authIds["Scherbeck, Christoph"]!!
        }
        BookAuthors.insert {
            it[bookId] = bkid
            it[authorId] = authIds["Kühnast, Charly"]!!
        }

        // Docker-Buch
        bkid = Books.insert {
            it[title] = "Docker -- Das Praxisbuch, 2. Aufl."
            it[publyear] = 2020
            it[publId] = publIds["Rheinwerk Verlag"]!!
        }[Books.id]
        BookAuthors.insert {
            it[bookId] = bkid
            it[authorId] = authIds["Öggl, Bernd"]!!
        }
        BookAuthors.insert {
            it[bookId] = bkid
            it[authorId] = authIds["Kofler, Michael"]!!
        }

        // Ubuntu-E-Book
        bkid = Books.insert {
            it[title] = "Ubuntu 20.04"
            it[publyear] = 2020
            it[publId] = publIds["ebooks.kofler"]!!
        }[Books.id]
        BookAuthors.insert {
            it[bookId] = bkid
            it[authorId] = authIds["Kofler, Michael"]!!
        }
        // Pandoc-E-Book
        bkid = Books.insert {
            it[title] = "Markdown & Pandoc"
            it[publyear] = 2019
            it[publId] = publIds["ebooks.kofler"]!!
        }[Books.id]
        BookAuthors.insert {
            it[bookId] = bkid
            it[authorId] = authIds["Kofler, Michael"]!!
        }
    }
}

fun findOrInsertPublisher(publname: String) =
    transaction  {
        val result = Publisher.find { Publishers.name.eq(publname) }
        if (result.count() > 0)
            result.first().id
        else
            Publishers.insert {
                it[name] = publname
            }[Publishers.id]
    }


/*
// gleichwertiger Code
fun findOrInsertPublisher1(publname: String) : EntityID<Int> {
    return transaction {
        val result = Publisher.find { Publishers.name.eq(publname) }
        return@transaction if (result.count() > 0)
            result.first().id
        else
            Publishers.insert {
                it[name] = publname
            }[Publishers.id]
    }
}
*/

fun queryBooks(isMySQL: Boolean) {
      transaction {
          addLogger(StdOutSqlLogger)

          // Schleife über alle Bücher (sortiert)
          val allbooks = Book.all().sortedBy { it.title }
          for (b in allbooks) {
              println("${b.title} (${b.publyear})")
              println("  Verlag:  ${b.publisher.name}")
              val authors = b.authors
                  .sortedBy { it.name }
                  .joinToString(separator = "; ") { it.name }
              println("  Autoren: $authors")
          }
          println("----")

          // Schleife über alle Verlage mit ihren Büchern
          val allpublishers = Publisher.all().sortedBy { it.name.lowercase() }
          for (p in allpublishers) {
              print("Verlag: ${p.name} ")
              println("hat ${p.books.count()} Bücher veröffentlicht")
          }
          println("----")

          // gezielter Zugriff auf bestimmte Datensätze
          // match setzt bei MySQL einen Volltextindex für die betreffende
          // Spalte voraus und hat dann auch eine andere Syntax, findet
          // nur vollständige Wörter
          println("match")
          if (isMySQL) {
              var somebooks = Book.find { Books.title.match("Handbuch Linux") }
              for (b in somebooks)
                  println("${b.title} (${b.publyear})")
          } else {
              var somebooks = Book.find { Books.title.match("%buch%") }
              for (b in somebooks)
                  println("${b.title} (${b.publyear})")
          }
          println("----")
          var morebooks = Book.find { Books.publyear.eq(2020) }
          for (b in morebooks)
              println("${b.title} (${b.publyear})")
          println("----")


          // gezielter Zugriff auf einen Datensatz
          var onebook = Publisher.findById(2)
          if (onebook is Publisher)
              println("Verlag: ${onebook.name}")

          val result = Books.selectAll()

    }
}
