import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.runBlocking
import org.jetbrains.exposed.dao.flushCache
import org.jetbrains.exposed.sql.*
import org.jetbrains.exposed.sql.transactions.TransactionManager
import org.jetbrains.exposed.sql.transactions.experimental.suspendedTransactionAsync
import org.jetbrains.exposed.sql.transactions.transaction
import java.io.File
import java.lang.Exception
import java.sql.Connection

fun main() {
    // Datenbank löschen, jedesmal neu starten
    File("my.db").delete()
    File("myh2.mv.db").delete()

    // H2
    Database.connect("jdbc:h2:./myh2", "org.h2.Driver")

    // oder SQLite
    // Database.connect("jdbc:sqlite:my.db?foreign_keys=on", "org.sqlite.JDBC")

    // oder MySQL. Setzt voraus, dass es die Datenbank test2 und den user test2 gibt,
    // der darauf zugreifen darf.
    // val db = Database.connect("jdbc:mysql://localhost/test2",
    //     driver = "com.mysql.cj.jdbc.Driver",
    //     user = "test2",
    //    password = "geheim")

    // Tabellen und Testdatensätze für das Books-Beispiel erzeugen
    setupBooks()

    // DSL-Syntax ausprobieren
    testSelect()
    // testSelect2()
    // testUpdateInsertDelete()
    // testAsync()
    // testSQL()


}

// verschiedene select-Varianten testen
fun testSelect() {
    transaction {
        // Logging
        addLogger(StdOutSqlLogger)

        val allBooks = Books.selectAll()    // Datentyp Query
        for (row in allBooks) {             // row: Datentyp ResultRow
            val title = row[Books.title]    // String
            val year = row[Books.publyear]  // Int
            println("Titel: $title ($year)")
        }
        println("---")
        val publ2020 = Books.select {
            Books.publyear.eq(2020)
        }
        println("${publ2020.count()} Datensätze gefunden")
        for (row in publ2020) {             // row: Datentyp ResultRow
            val title = row[Books.title]    // String
            val year = row[Books.publyear]  // Int
            println("Titel: $title ($year)")
        }

        println("---")
        val booksAndPubl = (Books innerJoin Publishers)
            .slice(Books.title, Books.publyear, Publishers.name)
            .selectAll()
            .orderBy(Books.publyear to SortOrder.ASC)
            .orderBy(Books.title)
        for (row in booksAndPubl) {   // row: Datentyp ResultRow
            println(row)
        }

        println("---")
        val booksAndAuthors = (Books innerJoin BookAuthors innerJoin Authors)
            .slice(Books.title, Authors.name.count())
            .selectAll()
            .groupBy(Books.id)
        for (row in booksAndAuthors) {   // row: Datentyp ResultRow
            val title = row[Books.title]
            val noOfAuthors = row[Authors.name.count()]
            println("$title wurde von $noOfAuthors Autor/Autoren verfasst")
        }
    }
}

// select-Ergebnisse weiterverarbeiten
fun testSelect2() {
    transaction {
        // Logging
        addLogger(StdOutSqlLogger)


        val myid = 123
        // Ergebnisdatentyp: String?
        val title = Books
            .slice(Books.title)
            .select {Books.id.eq(myid)}
            .firstOrNull()
            ?.get(Books.title)
        println("Buchtitel: $title")

        val publId = 2
        // Ergebnisdatentyp: List<String>
        val titles = Books
            .slice(Books.title)
            .select { Books.publId.eq(publId) }
            .map { it[Books.title]}
        println("Buchtitel des Verlags mit publID==2: $titles")

        val publMap = Publishers
            .slice(Publishers.id, Publishers.name)
            .selectAll()
            .associate { it[Publishers.id].value to it[Publishers.name] }
        println(publMap)
        println(publMap[2])
    }
}

fun testUpdateInsertDelete() {
    transaction {
        // Logging
        addLogger(StdOutSqlLogger)

        println("--- insert ---")
        val publId = Publishers.insertAndGetId { publisher ->
            publisher[name] = "O'Reilly"
        }
        val bookId = Books.insertAndGetId {book ->
            book[title] = "High Performance MySQL"
            book[publyear] = 2012
            book[Books.publId] = publId
        }
        for (authName in listOf("Schwartz", "Zaitev", "Tkachenko")) {
            val authId = Authors.insertAndGetId { author ->
                author[name] = authName
            }
            BookAuthors.insert { ba ->
                ba[BookAuthors.bookId] = bookId
                ba[authorId] = authId
            }
        }

        println("--- update ---")
        Books.update(
            { Books.title eq "High Performance MySQL" })
        { book ->
            book[Books.publyear] = 2012
        }


        println("--- delete ---")
        try {
            val n = Books.deleteWhere {
                Books.title eq "High Performance MySQL" }
            println("$n Datensätze gelöscht")
        } catch (e: Exception) {
            println("Kann wegen FK-Regeln nicht gelöscht werden. $e")
        }

    }
}

fun testAsync() {
    runBlocking {
        suspendedTransactionAsync(Dispatchers.IO) {
            val publId = Publishers.insertAndGetId { publisher ->
                publisher[name] = "O'Reilly"
            }
            val bookId = Books.insertAndGetId {book ->
                book[title] = "High Performance MySQL"
                book[publyear] = 2012
                book[Books.publId] = publId
            }
            for (authName in listOf("Schwartz", "Zaitev", "Tkachenko")) {
                val authId = Authors.insertAndGetId { author ->
                    author[name] = authName
                }
                BookAuthors.insert { ba ->
                    ba[BookAuthors.bookId] = bookId
                    ba[authorId] = authId
                }
            }
            println("done")
        }
    }
    Thread.sleep(1000L)
}


fun testSQL() {
    transaction {
        // Logging
        addLogger(StdOutSqlLogger)

        val selectCmd = (Books innerJoin BookAuthors innerJoin Authors)
            .slice(Books.title, Authors.name.count())
            .selectAll()
            .groupBy(Books.id).prepareSQL(QueryBuilder(false))
        println(selectCmd)

        val sql = "SELECT * FROM Books"
        exec(sql) {rs ->  // JDBC-ResultSet
            while (rs.next()) {
                val id = rs.getInt("id")
                val title = rs.getString("title")
                println("ID=$id: $title")
            }
        }

    }
}

